@@ -124,9 +124,12 @@ class ConvRecurrent2D(Recurrent):
     self.return_sequences = return_sequences
     self.go_backwards = go_backwards
     self.stateful = stateful
-    self.input_spec = InputSpec(ndim=5)
+    self.input_spec = [InputSpec(ndim=5)]
+    self.state_spec = None
 
   def _compute_output_shape(self, input_shape):
+    if isinstance(input_shape, list):
+      input_shape = input_shape[0]
     input_shape = tensor_shape.TensorShape(input_shape).as_list()
     if self.data_format == 'channels_first':
       rows = input_shape[3]
@@ -344,11 +347,14 @@ class ConvLSTM2D(ConvRecurrent2D):
 
     self.dropout = min(1., max(0., dropout))
     self.recurrent_dropout = min(1., max(0., recurrent_dropout))
+    self.state_spec = [InputSpec(ndim=4), InputSpec(ndim=4)]
 
   def build(self, input_shape):
-    input_shape = tensor_shape.TensorShape(input_shape).as_list()
-    # TODO(fchollet): better handling of input spec
-    self.input_spec = InputSpec(shape=input_shape)
+    if isinstance(input_shape, list):
+      input_shape = input_shape[0]
+    input_shape = tuple(tensor_shape.TensorShape(input_shape).as_list())
+    batch_size = input_shape[0] if self.stateful else None
+    self.input_spec[0] = InputSpec(shape=(batch_size, None) + input_shape[2:])
 
     if self.stateful:
       self.reset_states()
@@ -357,32 +363,39 @@ class ConvLSTM2D(ConvRecurrent2D):
       self.states = [None, None]
 
     if self.data_format == 'channels_first':
-      channel_axis = 1
+      channel_axis = 2
     else:
       channel_axis = -1
     if input_shape[channel_axis] is None:
       raise ValueError('The channel dimension of the inputs '
                        'should be defined. Found `None`.')
     input_dim = input_shape[channel_axis]
+    state_shape = [None] * 4
+    state_shape[channel_axis] = input_dim
+    state_shape = tuple(state_shape)
+    self.state_spec = [
+        InputSpec(shape=state_shape),
+        InputSpec(shape=state_shape)
+    ]
     kernel_shape = self.kernel_size + (input_dim, self.filters * 4)
     self.kernel_shape = kernel_shape
     recurrent_kernel_shape = self.kernel_size + (self.filters, self.filters * 4)
 
     self.kernel = self.add_weight(
-        kernel_shape,
+        shape=kernel_shape,
         initializer=self.kernel_initializer,
         name='kernel',
         regularizer=self.kernel_regularizer,
         constraint=self.kernel_constraint)
     self.recurrent_kernel = self.add_weight(
-        recurrent_kernel_shape,
+        shape=recurrent_kernel_shape,
         initializer=self.recurrent_initializer,
         name='recurrent_kernel',
         regularizer=self.recurrent_regularizer,
         constraint=self.recurrent_constraint)
     if self.use_bias:
       self.bias = self.add_weight(
-          (self.filters * 4,),
+          shape=(self.filters * 4,),
           initializer=self.bias_initializer,
           name='bias',
           regularizer=self.bias_regularizer,
@@ -417,7 +430,7 @@ class ConvLSTM2D(ConvRecurrent2D):
       self.bias_o = None
     self.built = True
 
-  def get_initial_states(self, inputs):
+  def get_initial_state(self, inputs):
     # (samples, timesteps, rows, cols, filters)
     initial_state = K.zeros_like(inputs)
     # (samples, rows, cols, filters)
@@ -433,8 +446,9 @@ class ConvLSTM2D(ConvRecurrent2D):
   def reset_states(self):
     if not self.stateful:
       raise RuntimeError('Layer must be stateful.')
-    input_shape = self.input_spec.shape
+    input_shape = self.input_spec[0].shape
     output_shape = self._compute_output_shape(input_shape)
+
     if not input_shape[0]:
       raise ValueError('If a RNN is stateful, a complete '
                        'input_shape must be provided '
@@ -453,8 +467,8 @@ class ConvLSTM2D(ConvRecurrent2D):
                   np.zeros((input_shape[0], out_row, out_col, out_filter)))
     else:
       self.states = [
-          K.zeros((input_shape[0], out_row, out_col, out_filter)), K.zeros(
-              (input_shape[0], out_row, out_col, out_filter))
+          K.zeros((input_shape[0], out_row, out_col, out_filter)),
+          K.zeros((input_shape[0], out_row, out_col, out_filter))
       ]
 
   def get_constants(self, inputs, training=None):
