@@ -19,17 +19,15 @@ from __future__ import print_function
 
 import math
 import random
-import sys
 
-from tensorflow.contrib.losses.python.losses import loss_ops
+from tensorflow.contrib.framework.python.ops import variables as framework_variables
 from tensorflow.contrib.tensor_forest.python import constants
-from tensorflow.contrib.tensor_forest.python.ops import inference_ops
-from tensorflow.contrib.tensor_forest.python.ops import training_ops
+from tensorflow.contrib.tensor_forest.python.ops import data_ops
+from tensorflow.contrib.tensor_forest.python.ops import tensor_forest_ops
 
 from tensorflow.python.framework import constant_op
 from tensorflow.python.framework import dtypes
 from tensorflow.python.framework import ops
-from tensorflow.python.framework import sparse_tensor
 from tensorflow.python.ops import array_ops
 from tensorflow.python.ops import control_flow_ops
 from tensorflow.python.ops import init_ops
@@ -66,7 +64,7 @@ class ForestHParams(object):
                split_after_samples=250,
                min_split_samples=5,
                valid_leaf_threshold=1,
-               dominate_method='hoeffding',
+               dominate_method='bootstrap',
                dominate_fraction=0.99,
                **kwargs):
     self.num_trees = num_trees
@@ -112,12 +110,10 @@ class ForestHParams(object):
     # regression and avoids having to recompute sums for classification.
     self.num_output_columns = self.num_classes + 1
 
-    # The Random Forest literature recommends sqrt(# features) for
-    # classification problems, and p/3 for regression problems.
-    # TODO(thomaswc): Consider capping this for large number of features.
-    self.num_splits_to_consider = (
-        self.num_splits_to_consider or
-        max(10, int(math.ceil(math.sqrt(self.num_features)))))
+    # Our experiments have found that num_splits_to_consider = num_features
+    # gives good accuracy.
+    self.num_splits_to_consider = self.num_splits_to_consider or min(
+        self.num_features, 1000)
 
     self.max_fertile_nodes = (self.max_fertile_nodes or
                               int(math.ceil(self.max_nodes / 2.0)))
@@ -283,8 +279,16 @@ class ForestTrainingVariables(object):
   def __init__(self, params, device_assigner, training=True,
                tree_variables_class=TreeTrainingVariables):
     self.variables = []
+    # Set up some scalar variables to run through the device assigner, then
+    # we can use those to colocate everything related to a tree.
+    self.device_dummies = []
+    with ops.device(device_assigner):
+      for i in range(params.num_trees):
+        self.device_dummies.append(variable_scope.get_variable(
+            name='device_dummy_%d' % i, shape=0))
+
     for i in range(params.num_trees):
-      with ops.device(device_assigner.get_device(i)):
+      with ops.device(self.device_dummies[i].device):
         self.variables.append(tree_variables_class(params, i, training))
 
   def __setitem__(self, t, val):
@@ -294,35 +298,19 @@ class ForestTrainingVariables(object):
     return self.variables[t]
 
 
-class RandomForestDeviceAssigner(object):
-  """A device assigner that uses the default device.
-
-  Write subclasses that implement get_device for control over how trees
-  get assigned to devices.  This assumes that whole trees are assigned
-  to a device.
-  """
-
-  def __init__(self):
-    self.cached = None
-
-  def get_device(self, unused_tree_num):
-    if not self.cached:
-      dummy = constant_op.constant(0)
-      self.cached = dummy.device
-
-    return self.cached
-
-
 class RandomForestGraphs(object):
   """Builds TF graphs for random forest training and inference."""
 
-  def __init__(self, params, device_assigner=None,
-               variables=None, tree_variables_class=TreeTrainingVariables,
-               tree_graphs=None, training=True,
-               t_ops=training_ops,
-               i_ops=inference_ops):
+  def __init__(self,
+               params,
+               device_assigner=None,
+               variables=None,
+               tree_variables_class=TreeTrainingVariables,
+               tree_graphs=None,
+               training=True):
     self.params = params
-    self.device_assigner = device_assigner or RandomForestDeviceAssigner()
+    self.device_assigner = (
+        device_assigner or framework_variables.VariableDeviceChooser())
     logging.info('Constructing forest with params = ')
     logging.info(self.params.__dict__)
     self.variables = variables or ForestTrainingVariables(
@@ -330,48 +318,68 @@ class RandomForestGraphs(object):
         tree_variables_class=tree_variables_class)
     tree_graph_class = tree_graphs or RandomTreeGraphs
     self.trees = [
-        tree_graph_class(
-            self.variables[i], self.params,
-            t_ops.Load(), i_ops.Load(), i)
-        for i in range(self.params.num_trees)]
+        tree_graph_class(self.variables[i], self.params, i)
+        for i in range(self.params.num_trees)
+    ]
 
   def _bag_features(self, tree_num, input_data):
-    split_data = array_ops.split(1, self.params.num_features, input_data)
+    split_data = array_ops.split(
+        value=input_data, num_or_size_splits=self.params.num_features, axis=1)
     return array_ops.concat(
-        1, [split_data[ind] for ind in self.params.bagged_features[tree_num]])
+        [split_data[ind] for ind in self.params.bagged_features[tree_num]], 1)
 
   def training_graph(self,
                      input_data,
                      input_labels,
-                     data_spec=None,
+                     num_trainers=1,
+                     trainer_id=0,
                      **tree_kwargs):
     """Constructs a TF graph for training a random forest.
 
     Args:
-      input_data: A tensor or SparseTensor or placeholder for input data.
+      input_data: A tensor or dict of string->Tensor for input data.
       input_labels: A tensor or placeholder for labels associated with
         input_data.
-      data_spec: A list of tf.dtype values specifying the original types of
-        each column.
+      num_trainers: Number of parallel trainers to split trees among.
+      trainer_id: Which trainer this instance is.
       **tree_kwargs: Keyword arguments passed to each tree's training_graph.
 
     Returns:
       The last op in the random forest training graph.
+
+    Raises:
+      NotImplementedError: If trying to use bagging with sparse features.
     """
-    data_spec = [constants.DATA_FLOAT] if data_spec is None else data_spec
+    processed_dense_features, processed_sparse_features, data_spec = (
+        data_ops.ParseDataTensorOrDict(input_data))
+
+    if input_labels is not None:
+      labels = data_ops.ParseLabelTensorOrDict(input_labels)
+
+    data_spec = data_spec or self.get_default_data_spec(input_data)
+
     tree_graphs = []
-    for i in range(self.params.num_trees):
-      with ops.device(self.device_assigner.get_device(i)):
+    trees_per_trainer = self.params.num_trees / num_trainers
+    tree_start = int(trainer_id * trees_per_trainer)
+    tree_end = int((trainer_id + 1) * trees_per_trainer)
+    for i in range(tree_start, tree_end):
+      logging.info('training graph for tree: %d' % i)
+      with ops.device(self.variables.device_dummies[i].device):
         seed = self.params.base_random_seed
         if seed != 0:
           seed += i
         # If using bagging, randomly select some of the input.
-        tree_data = input_data
-        tree_labels = input_labels
+        tree_data = processed_dense_features
+        tree_labels = labels
         if self.params.bagging_fraction < 1.0:
+          # TODO(gilberth): Support bagging for sparse features.
+          if processed_sparse_features is not None:
+            raise NotImplementedError(
+                'Bagging not supported with sparse features.')
           # TODO(thomaswc): This does sampling without replacment.  Consider
           # also allowing sampling with replacement as an option.
-          batch_size = array_ops.slice(array_ops.shape(input_data), [0], [1])
+          batch_size = array_ops.strided_slice(
+              array_ops.shape(processed_dense_features), [0], [1])
           r = random_ops.random_uniform(batch_size, seed=seed)
           mask = math_ops.less(
               r, array_ops.ones_like(r) * self.params.bagging_fraction)
@@ -379,44 +387,60 @@ class RandomForestGraphs(object):
               array_ops.where(mask), squeeze_dims=[1])
           # TODO(thomaswc): Calculate out-of-bag data and labels, and store
           # them for use in calculating statistics later.
-          tree_data = array_ops.gather(input_data, gather_indices)
-          tree_labels = array_ops.gather(input_labels, gather_indices)
+          tree_data = array_ops.gather(processed_dense_features, gather_indices)
+          tree_labels = array_ops.gather(labels, gather_indices)
         if self.params.bagged_features:
+          if processed_sparse_features is not None:
+            raise NotImplementedError(
+                'Feature bagging not supported with sparse features.')
           tree_data = self._bag_features(i, tree_data)
 
         initialization = self.trees[i].tree_initialization()
 
         with ops.control_dependencies([initialization]):
-          tree_graphs.append(
-              self.trees[i].training_graph(
-                  tree_data, tree_labels, seed, data_spec=data_spec,
-                  **tree_kwargs))
+          tree_graphs.append(self.trees[i].training_graph(
+              tree_data,
+              tree_labels,
+              seed,
+              data_spec=data_spec,
+              sparse_features=processed_sparse_features,
+              **tree_kwargs))
 
     return control_flow_ops.group(*tree_graphs, name='train')
 
-  def inference_graph(self, input_data, data_spec=None, **inference_args):
+  def inference_graph(self, input_data, **inference_args):
     """Constructs a TF graph for evaluating a random forest.
 
     Args:
-      input_data: A tensor or SparseTensor or placeholder for input data.
-      data_spec: A list of tf.dtype values specifying the original types of
-        each column.
+      input_data: A tensor or dict of string->Tensor for input data.
       **inference_args: Keyword arguments to pass through to each tree.
 
     Returns:
       The last op in the random forest inference graph.
+
+    Raises:
+      NotImplementedError: If trying to use feature bagging with sparse
+        features.
     """
-    data_spec = [constants.DATA_FLOAT] if data_spec is None else data_spec
+    processed_dense_features, processed_sparse_features, data_spec = (
+        data_ops.ParseDataTensorOrDict(input_data))
+
     probabilities = []
     for i in range(self.params.num_trees):
-      with ops.device(self.device_assigner.get_device(i)):
-        tree_data = input_data
+      with ops.device(self.variables.device_dummies[i].device):
+        tree_data = processed_dense_features
         if self.params.bagged_features:
-          tree_data = self._bag_features(i, input_data)
+          if processed_sparse_features is not None:
+            raise NotImplementedError(
+                'Feature bagging not supported with sparse features.')
+          tree_data = self._bag_features(i, tree_data)
         probabilities.append(self.trees[i].inference_graph(
-            tree_data, data_spec, **inference_args))
-    with ops.device(self.device_assigner.get_device(0)):
-      all_predict = array_ops.pack(probabilities)
+            tree_data,
+            data_spec,
+            sparse_features=processed_sparse_features,
+            **inference_args))
+    with ops.device(self.variables.device_dummies[0].device):
+      all_predict = array_ops.stack(probabilities)
       return math_ops.div(
           math_ops.reduce_sum(all_predict, 0), self.params.num_trees,
           name='probabilities')
@@ -429,18 +453,17 @@ class RandomForestGraphs(object):
     """
     sizes = []
     for i in range(self.params.num_trees):
-      with ops.device(self.device_assigner.get_device(i)):
+      with ops.device(self.variables.device_dummies[i].device):
         sizes.append(self.trees[i].size())
-    return math_ops.reduce_mean(math_ops.to_float(array_ops.pack(sizes)))
+    return math_ops.reduce_mean(math_ops.to_float(array_ops.stack(sizes)))
 
   # pylint: disable=unused-argument
-  def training_loss(self, features, labels, data_spec=None,
-                    name='training_loss'):
-    return math_ops.neg(self.average_size(), name=name)
+  def training_loss(self, features, labels, name='training_loss'):
+    return math_ops.negative(self.average_size(), name=name)
 
   # pylint: disable=unused-argument
   def validation_loss(self, features, labels):
-    return math_ops.neg(self.average_size())
+    return math_ops.negative(self.average_size())
 
   def average_impurity(self):
     """Constructs a TF graph for evaluating the leaf impurity of a forest.
@@ -450,81 +473,43 @@ class RandomForestGraphs(object):
     """
     impurities = []
     for i in range(self.params.num_trees):
-      with ops.device(self.device_assigner.get_device(i)):
+      with ops.device(self.variables.device_dummies[i].device):
         impurities.append(self.trees[i].average_impurity())
-    return math_ops.reduce_mean(array_ops.pack(impurities))
+    return math_ops.reduce_mean(array_ops.stack(impurities))
 
   def get_stats(self, session):
     tree_stats = []
     for i in range(self.params.num_trees):
-      with ops.device(self.device_assigner.get_device(i)):
+      with ops.device(self.variables.device_dummies[i].device):
         tree_stats.append(self.trees[i].get_stats(session))
     return ForestStats(tree_stats, self.params)
 
+  def feature_importances(self):
+    tree_counts = [self.trees[i].feature_usage_counts()
+                   for i in range(self.params.num_trees)]
+    total_counts = math_ops.reduce_sum(array_ops.stack(tree_counts, 0), 0)
+    return total_counts / math_ops.reduce_sum(total_counts)
+
 
 def one_hot_wrapper(num_classes, loss_fn):
   """Some loss functions take one-hot labels."""
   def _loss(probs, targets):
+    if targets.get_shape().ndims > 1:
+      targets = array_ops.squeeze(targets, squeeze_dims=[1])
     one_hot_labels = array_ops.one_hot(
-        math_ops.to_int32(targets), num_classes,
-        on_value=1., off_value=0., dtype=dtypes.float32)
+        math_ops.to_int32(targets),
+        num_classes,
+        on_value=1.,
+        off_value=0.,
+        dtype=dtypes.float32)
     return loss_fn(probs, one_hot_labels)
   return _loss
 
 
-class TrainingLossForest(RandomForestGraphs):
-  """Random Forest that uses training loss as the termination criteria."""
-
-  def __init__(self, params, loss_fn=None, **kwargs):
-    """Initialize.
-
-    Args:
-      params: Like RandomForestGraphs, a ForestHParams object.
-      loss_fn: A function that takes probabilities and targets and returns
-        a loss for each example.
-      **kwargs: Keyword args to pass to superclass (RandomForestGraphs).
-    """
-    self.loss_fn = loss_fn or one_hot_wrapper(params.num_classes,
-                                              loss_ops.log_loss)
-    self._loss = None
-    super(TrainingLossForest, self).__init__(params, **kwargs)
-
-  def _get_loss(self, features, labels, data_spec=None):
-    """Constructs, caches, and returns the inference-based loss."""
-    if self._loss is not None:
-      return self._loss
-
-    def _average_loss():
-      probs = self.inference_graph(features, data_spec=data_spec)
-      return math_ops.reduce_sum(self.loss_fn(
-          probs, labels)) / math_ops.to_float(
-              array_ops.shape(features)[0])
-
-    self._loss = control_flow_ops.cond(
-        self.average_size() > 0, _average_loss,
-        lambda: constant_op.constant(sys.maxsize, dtype=dtypes.float32))
-
-    return self._loss
-
-  def training_graph(self, input_data, input_labels, data_spec=None,
-                     **kwargs):
-    loss = self._get_loss(input_data, input_labels, data_spec=data_spec)
-    with ops.control_dependencies([loss.op]):
-      return super(TrainingLossForest, self).training_graph(
-          input_data, input_labels, **kwargs)
-
-  def training_loss(self, features, labels, data_spec=None,
-                    name='training_loss'):
-    return array_ops.identity(
-        self._get_loss(features, labels, data_spec=data_spec), name=name)
-
-
 class RandomTreeGraphs(object):
   """Builds TF graphs for random tree training and inference."""
 
-  def __init__(self, variables, params, t_ops, i_ops, tree_num):
-    self.training_ops = t_ops
-    self.inference_ops = i_ops
+  def __init__(self, variables, params, tree_num):
     self.variables = variables
     self.params = params
     self.tree_num = tree_num
@@ -537,9 +522,10 @@ class RandomTreeGraphs(object):
       return control_flow_ops.no_op()
 
     return control_flow_ops.cond(
-        math_ops.equal(array_ops.squeeze(array_ops.slice(
-            self.variables.tree, [0, 0], [1, 1])), -2),
-        _init_tree, _nothing)
+        math_ops.equal(
+            array_ops.squeeze(
+                array_ops.strided_slice(self.variables.tree, [0, 0], [1, 1])),
+            -2), _init_tree, _nothing)
 
   def _gini(self, class_counts):
     """Calculate the Gini impurity.
@@ -604,18 +590,20 @@ class RandomTreeGraphs(object):
                      input_labels,
                      random_seed,
                      data_spec,
+                     sparse_features=None,
                      input_weights=None):
 
     """Constructs a TF graph for training a random tree.
 
     Args:
-      input_data: A tensor or SparseTensor or placeholder for input data.
+      input_data: A tensor or placeholder for input data.
       input_labels: A tensor or placeholder for labels associated with
         input_data.
       random_seed: The random number generator seed to use for this tree.  0
         means use the current time as the seed.
-      data_spec: A list of tf.dtype values specifying the original types of
-        each column.
+      data_spec: A data_ops.TensorForestDataSpec object specifying the
+        original feature/columns of the data.
+      sparse_features: A tf.SparseTensor for sparse input data.
       input_weights: A float tensor or placeholder holding per-input weights,
         or None if all inputs are to be weighted equally.
 
@@ -624,27 +612,30 @@ class RandomTreeGraphs(object):
     """
     epoch = math_ops.to_int32(get_epoch_variable())
 
+    serialized_input_spec = data_spec.SerializeToString()
+
     if input_weights is None:
       input_weights = []
 
+    if input_data is None:
+      input_data = []
+
     sparse_indices = []
     sparse_values = []
     sparse_shape = []
-    if isinstance(input_data, sparse_tensor.SparseTensor):
-      sparse_indices = input_data.indices
-      sparse_values = input_data.values
-      sparse_shape = input_data.shape
-      input_data = []
+    if sparse_features is not None:
+      sparse_indices = sparse_features.indices
+      sparse_values = sparse_features.values
+      sparse_shape = sparse_features.dense_shape
 
     # Count extremely random stats.
     (node_sums, node_squares, splits_indices, splits_sums, splits_squares,
      totals_indices, totals_sums, totals_squares,
-     input_leaves) = (self.training_ops.count_extremely_random_stats(
+     input_leaves) = (tensor_forest_ops.count_extremely_random_stats(
          input_data,
          sparse_indices,
          sparse_values,
          sparse_shape,
-         data_spec,
          input_labels,
          input_weights,
          self.variables.tree,
@@ -654,6 +645,7 @@ class RandomTreeGraphs(object):
          self.variables.candidate_split_thresholds,
          self.variables.start_epoch,
          epoch,
+         input_spec=serialized_input_spec,
          num_classes=self.params.num_output_columns,
          regression=self.params.regression))
     node_update_ops = []
@@ -661,26 +653,27 @@ class RandomTreeGraphs(object):
         state_ops.assign_add(self.variables.node_sums, node_sums))
 
     splits_update_ops = []
-    splits_update_ops.append(self.training_ops.scatter_add_ndim(
-        self.variables.candidate_split_sums,
-        splits_indices, splits_sums))
-    splits_update_ops.append(self.training_ops.scatter_add_ndim(
-        self.variables.accumulator_sums, totals_indices,
-        totals_sums))
+    splits_update_ops.append(
+        tensor_forest_ops.scatter_add_ndim(self.variables.candidate_split_sums,
+                                           splits_indices, splits_sums))
+    splits_update_ops.append(
+        tensor_forest_ops.scatter_add_ndim(self.variables.accumulator_sums,
+                                           totals_indices, totals_sums))
 
     if self.params.regression:
       node_update_ops.append(state_ops.assign_add(self.variables.node_squares,
                                                   node_squares))
-      splits_update_ops.append(self.training_ops.scatter_add_ndim(
-          self.variables.candidate_split_squares,
-          splits_indices, splits_squares))
-      splits_update_ops.append(self.training_ops.scatter_add_ndim(
-          self.variables.accumulator_squares, totals_indices,
-          totals_squares))
+      splits_update_ops.append(
+          tensor_forest_ops.scatter_add_ndim(
+              self.variables.candidate_split_squares, splits_indices,
+              splits_squares))
+      splits_update_ops.append(
+          tensor_forest_ops.scatter_add_ndim(self.variables.accumulator_squares,
+                                             totals_indices, totals_squares))
 
     # Sample inputs.
     update_indices, feature_updates, threshold_updates = (
-        self.training_ops.sample_inputs(
+        tensor_forest_ops.sample_inputs(
             input_data,
             sparse_indices,
             sparse_values,
@@ -690,6 +683,7 @@ class RandomTreeGraphs(object):
             input_leaves,
             self.variables.candidate_split_features,
             self.variables.candidate_split_thresholds,
+            input_spec=serialized_input_spec,
             split_initializations_per_input=(
                 self.params.split_initializations_per_input),
             split_sampling_random_seed=random_seed))
@@ -706,7 +700,7 @@ class RandomTreeGraphs(object):
       # have become stale won't be deallocated until an input reaches them,
       # because we're trying to avoid considering every fertile node for
       # performance reasons.
-      finished, stale = self.training_ops.finished_nodes(
+      finished, stale = tensor_forest_ops.finished_nodes(
           input_leaves,
           self.variables.node_to_accumulator_map,
           self.variables.candidate_split_sums,
@@ -747,8 +741,9 @@ class RandomTreeGraphs(object):
 
     # Calculate best splits.
     with ops.control_dependencies(splits_update_ops):
-      split_indices = self.training_ops.best_splits(
-          finished, self.variables.node_to_accumulator_map,
+      split_indices = tensor_forest_ops.best_splits(
+          finished,
+          self.variables.node_to_accumulator_map,
           self.variables.candidate_split_sums,
           self.variables.candidate_split_squares,
           self.variables.accumulator_sums,
@@ -756,9 +751,10 @@ class RandomTreeGraphs(object):
           regression=self.params.regression)
 
     # Grow tree.
-    with ops.control_dependencies([update_features_op, update_thresholds_op]):
+    with ops.control_dependencies([update_features_op, update_thresholds_op,
+                                   non_fertile_leaves.op]):
       (tree_update_indices, tree_children_updates, tree_threshold_updates,
-       new_eot) = (self.training_ops.grow_tree(
+       new_eot) = (tensor_forest_ops.grow_tree(
            self.variables.end_of_tree, self.variables.node_to_accumulator_map,
            finished, split_indices, self.variables.candidate_split_features,
            self.variables.candidate_split_thresholds))
@@ -777,7 +773,7 @@ class RandomTreeGraphs(object):
     # Update fertile slots.
     with ops.control_dependencies([tree_update_op]):
       (n2a_map_updates, a2n_map_updates, accumulators_cleared,
-       accumulators_allocated) = (self.training_ops.update_fertile_slots(
+       accumulators_allocated) = (tensor_forest_ops.update_fertile_slots(
            finished,
            non_fertile_leaves,
            non_fertile_leaf_scores,
@@ -790,9 +786,8 @@ class RandomTreeGraphs(object):
 
     # Ensure end_of_tree doesn't get updated until UpdateFertileSlots has
     # used it to calculate new leaves.
-    gated_new_eot, = control_flow_ops.tuple(
-        [new_eot], control_inputs=[n2a_map_updates])
-    eot_update_op = state_ops.assign(self.variables.end_of_tree, gated_new_eot)
+    with ops.control_dependencies([n2a_map_updates.op]):
+      eot_update_op = state_ops.assign(self.variables.end_of_tree, new_eot)
 
     updates = []
     updates.append(eot_update_op)
@@ -809,7 +804,7 @@ class RandomTreeGraphs(object):
                                  a2n_map_updates[0], a2n_map_updates[1]))
 
     cleared_and_allocated_accumulators = array_ops.concat(
-        0, [accumulators_cleared, accumulators_allocated])
+        [accumulators_cleared, accumulators_allocated], 0)
 
     # Calculate values to put into scatter update for candidate counts.
     # Candidate split counts are always reset back to 0 for both cleared
@@ -831,15 +826,15 @@ class RandomTreeGraphs(object):
     # Calculate values to put into scatter update for total counts.
     total_cleared = array_ops.tile(
         array_ops.expand_dims(
-            math_ops.neg(array_ops.ones_like(accumulators_cleared,
-                                             dtype=dtypes.float32)), 1),
+            math_ops.negative(array_ops.ones_like(accumulators_cleared,
+                                                  dtype=dtypes.float32)), 1),
         [1, self.params.num_output_columns])
     total_reset = array_ops.tile(
         array_ops.expand_dims(
             array_ops.zeros_like(accumulators_allocated,
                                  dtype=dtypes.float32), 1),
         [1, self.params.num_output_columns])
-    accumulator_updates = array_ops.concat(0, [total_cleared, total_reset])
+    accumulator_updates = array_ops.concat([total_cleared, total_reset], 0)
     updates.append(state_ops.scatter_update(
         self.variables.accumulator_sums,
         cleared_and_allocated_accumulators, accumulator_updates))
@@ -851,7 +846,7 @@ class RandomTreeGraphs(object):
     # Calculate values to put into scatter update for candidate splits.
     split_features_updates = array_ops.tile(
         array_ops.expand_dims(
-            math_ops.neg(array_ops.ones_like(
+            math_ops.negative(array_ops.ones_like(
                 cleared_and_allocated_accumulators)), 1),
         [1, self.params.num_splits_to_consider])
     updates.append(state_ops.scatter_update(
@@ -873,30 +868,38 @@ class RandomTreeGraphs(object):
     """
     return []
 
-  def inference_graph(self, input_data, data_spec):
+  def inference_graph(self, input_data, data_spec, sparse_features=None):
     """Constructs a TF graph for evaluating a random tree.
 
     Args:
-      input_data: A tensor or SparseTensor or placeholder for input data.
-      data_spec: A list of tf.dtype values specifying the original types of
-        each column.
+      input_data: A tensor or placeholder for input data.
+      data_spec: A TensorForestDataSpec proto specifying the original
+        input columns.
+      sparse_features: A tf.SparseTensor for sparse input data.
 
     Returns:
       The last op in the random tree inference graph.
     """
+    if input_data is None:
+      input_data = []
+
     sparse_indices = []
     sparse_values = []
     sparse_shape = []
-    if isinstance(input_data, sparse_tensor.SparseTensor):
-      sparse_indices = input_data.indices
-      sparse_values = input_data.values
-      sparse_shape = input_data.shape
-      input_data = []
-    return self.inference_ops.tree_predictions(
-        input_data, sparse_indices, sparse_values, sparse_shape, data_spec,
+    if sparse_features is not None:
+      sparse_indices = sparse_features.indices
+      sparse_values = sparse_features.values
+      sparse_shape = sparse_features.dense_shape
+
+    return tensor_forest_ops.tree_predictions(
+        input_data,
+        sparse_indices,
+        sparse_values,
+        sparse_shape,
         self.variables.tree,
         self.variables.tree_thresholds,
         self.variables.node_sums,
+        input_spec=data_spec.SerializeToString(),
         valid_leaf_threshold=self.params.valid_leaf_threshold)
 
   def average_impurity(self):
@@ -940,3 +943,10 @@ class RandomTreeGraphs(object):
             self.variables.tree, [0, 0], [-1, 1])), constants.LEAF_NODE)
         ).eval(session=session).shape[0]
     return TreeStats(num_nodes, num_leaves)
+
+  def feature_usage_counts(self):
+    features = array_ops.slice(self.variables.tree, [0, 1], [-1, 1])
+    # One hot ignores negative values, which is the default for unused nodes.
+    one_hots = array_ops.one_hot(
+        array_ops.squeeze(features), self.params.num_features)
+    return math_ops.reduce_sum(one_hots, 0)
