@@ -27,6 +27,7 @@ import collections
 import copy
 import functools
 import re
+import weakref
 
 from six.moves import xrange  # pylint: disable=redefined-builtin
 import numpy as np
@@ -40,6 +41,14 @@ from tensorflow.python.util import nest
 from tensorflow.python.util import tf_inspect
 
 
+def _is_tensor_or_tensor_list(v):
+  v = nest.flatten(v)
+  if v and isinstance(v[0], ops.Tensor):
+    return True
+  else:
+    return False
+
+
 class Layer(object):
   """Base layer class.
 
@@ -472,6 +481,8 @@ class Layer(object):
         setattr(result, k, v)
       elif k in shallow_copy:
         setattr(result, k, copy.copy(v))
+      elif _is_tensor_or_tensor_list(v):
+        setattr(result, k, v)
       else:
         setattr(result, k, copy.deepcopy(v, memo))
     return result
@@ -671,8 +682,7 @@ def _object_list_uid(object_list):
 # A global dictionary mapping graph objects to an index of counters used
 # for various layer names in each graph.
 # Allows to give unique autogenerated names to layers, in a graph-specific way.
-PER_GRAPH_LAYER_NAME_UIDS = collections.defaultdict(
-    lambda: collections.defaultdict(int))
+PER_GRAPH_LAYER_NAME_UIDS = weakref.WeakKeyDictionary()
 
 
 def _unique_layer_name(name):
@@ -694,6 +704,8 @@ def _unique_layer_name(name):
   ```
   """
   graph = ops.get_default_graph()
+  if graph not in PER_GRAPH_LAYER_NAME_UIDS:
+    PER_GRAPH_LAYER_NAME_UIDS[graph] = collections.defaultdict(int)
   layer_name_uids = PER_GRAPH_LAYER_NAME_UIDS[graph]
   layer_name_uids[name] += 1
   return name + '_' + str(layer_name_uids[name])
