@@ -17,6 +17,9 @@
 
 See the @{$python/test} guide.
 
+Note: `tf.test.mock` is an alias to the python `mock` or `unittest.mock`
+depending on the python version.
+
 @@main
 @@TestCase
 @@test_src_dir_path
