@@ -20,11 +20,17 @@ from __future__ import print_function
 
 import collections
 import functools
-import inspect
 import re
 
 from tensorflow.python.platform import tf_logging as logging
 from tensorflow.python.util import decorator_utils
+from tensorflow.python.util import tf_contextlib
+from tensorflow.python.util import tf_decorator
+from tensorflow.python.util import tf_inspect
+
+
+# Allow deprecation warnings to be silenced temporarily with a context manager.
+_PRINT_DEPRECATION_WARNINGS = True
 
 
 def _add_deprecated_function_notice_to_docstring(doc, date, instructions):
@@ -33,7 +39,8 @@ def _add_deprecated_function_notice_to_docstring(doc, date, instructions):
       doc, instructions,
       'DEPRECATED FUNCTION',
       '(deprecated)', [
-          'THIS FUNCTION IS DEPRECATED. It will be removed after %s.' % date,
+          'THIS FUNCTION IS DEPRECATED. It will be removed %s.' % (
+              'in a future version' if date is None else ('after %s' % date)),
           'Instructions for updating:'])
 
 
@@ -44,14 +51,13 @@ def _add_deprecated_arg_notice_to_docstring(doc, date, instructions):
       'DEPRECATED FUNCTION ARGUMENTS',
       '(deprecated arguments)', [
           'SOME ARGUMENTS ARE DEPRECATED. '
-          'They will be removed after %s.' % date,
+          'They will be removed %s.' % (
+              'in a future version' if date is None else ('after %s' % date)),
           'Instructions for updating:'])
 
 
 def _validate_deprecation_args(date, instructions):
-  if not date:
-    raise ValueError('Tell us what date this will be deprecated!')
-  if not re.match(r'20\d\d-[01]\d-[0123]\d', date):
+  if date is not None and not re.match(r'20\d\d-[01]\d-[0123]\d', date):
     raise ValueError('Date must be YYYY-MM-DD.')
   if not instructions:
     raise ValueError('Don\'t deprecate things without conversion instructions!')
@@ -59,7 +65,7 @@ def _validate_deprecation_args(date, instructions):
 
 def _call_location():
   """Returns call location given level up from current call."""
-  frame = inspect.currentframe()
+  frame = tf_inspect.currentframe()
   if frame:
     # CPython internals are available, use them for performance.
     # walk back two frames to get to deprecated function caller.
@@ -69,7 +75,7 @@ def _call_location():
     return '%s:%d' % (frame.f_code.co_filename, frame.f_lineno)
   else:
     # Slow fallback path
-    stack = inspect.stack(0)  # 0 avoids generating unused context
+    stack = tf_inspect.stack(0)  # 0 avoids generating unused context
     entry = stack[2]
     return '%s:%d' % (entry[1], entry[2])
 
@@ -84,6 +90,7 @@ def deprecated(date, instructions):
     Instructions for updating:
     <instructions>
 
+  If `date` is None, 'after <date>' is replaced with 'in a future version'.
   <function> will include the class name if it is a method.
 
   It also edits the docstring of the function: ' (deprecated)' is appended
@@ -91,8 +98,8 @@ def deprecated(date, instructions):
   to the rest of the docstring.
 
   Args:
-    date: String. The date the function is scheduled to be removed. Must be
-      ISO 8601 (YYYY-MM-DD).
+    date: String or None. The date the function is scheduled to be removed.
+      Must be ISO 8601 (YYYY-MM-DD), or None.
     instructions: String. Instructions on how to update code using the
       deprecated function.
 
@@ -100,7 +107,8 @@ def deprecated(date, instructions):
     Decorated function or method.
 
   Raises:
-    ValueError: If date is not in ISO 8601 format, or instructions are empty.
+    ValueError: If date is not None or in ISO 8601 format, or instructions are
+      empty.
   """
   _validate_deprecation_args(date, instructions)
 
@@ -108,17 +116,20 @@ def deprecated(date, instructions):
     """Deprecation wrapper."""
     decorator_utils.validate_callable(func, 'deprecated')
     @functools.wraps(func)
-    def new_func(*args, **kwargs):
-      logging.warning(
-          'From %s: %s (from %s) is deprecated and will be removed '
-          'after %s.\n'
-          'Instructions for updating:\n%s',
-          _call_location(), decorator_utils.get_qualified_name(func),
-          func.__module__, date, instructions)
+    def new_func(*args, **kwargs):  # pylint: disable=missing-docstring
+      if _PRINT_DEPRECATION_WARNINGS:
+        logging.warning(
+            'From %s: %s (from %s) is deprecated and will be removed %s.\n'
+            'Instructions for updating:\n%s',
+            _call_location(), decorator_utils.get_qualified_name(func),
+            func.__module__,
+            'in a future version' if date is None else ('after %s' % date),
+            instructions)
       return func(*args, **kwargs)
-    new_func.__doc__ = _add_deprecated_function_notice_to_docstring(
-        func.__doc__, date, instructions)
-    return new_func
+    return tf_decorator.make_decorator(
+        func, new_func, 'deprecated',
+        _add_deprecated_function_notice_to_docstring(func.__doc__, date,
+                                                     instructions))
   return deprecated_wrapper
 
 
@@ -136,15 +147,16 @@ def deprecated_args(date, instructions, *deprecated_arg_names_or_tuples):
     removed after <date>. Instructions for updating:
       <instructions>
 
-  <function> will include the class name if it is a method.
+  If `date` is None, 'after <date>' is replaced with 'in a future version'.
+  <function> includes the class name if it is a method.
 
   It also edits the docstring of the function: ' (deprecated arguments)' is
   appended to the first line of the docstring and a deprecation notice is
   prepended to the rest of the docstring.
 
   Args:
-    date: String. The date the function is scheduled to be removed. Must be
-      ISO 8601 (YYYY-MM-DD).
+    date: String or None. The date the function is scheduled to be removed.
+      Must be ISO 8601 (YYYY-MM-DD), or None.
     instructions: String. Instructions on how to update code using the
       deprecated function.
     *deprecated_arg_names_or_tuples: String. or 2-Tuple(String,
@@ -156,7 +168,7 @@ def deprecated_args(date, instructions, *deprecated_arg_names_or_tuples):
     Decorated function or method.
 
   Raises:
-    ValueError: If date is not in ISO 8601 format, instructions are
+    ValueError: If date is not None or in ISO 8601 format, instructions are
       empty, the deprecated arguments are not present in the function
       signature, or the second element of a deprecated_tuple is not a
       list.
@@ -176,7 +188,7 @@ def deprecated_args(date, instructions, *deprecated_arg_names_or_tuples):
     return d
 
   def _get_deprecated_positional_arguments(names_to_ok_vals, arg_spec):
-    """Builds a dictionary from deprecated arguments to thier spec.
+    """Builds a dictionary from deprecated arguments to their spec.
 
     Returned dict is keyed by argument name.
     Each value is a DeprecatedArgSpec with the following fields:
@@ -189,7 +201,7 @@ def deprecated_args(date, instructions, *deprecated_arg_names_or_tuples):
     Args:
       names_to_ok_vals: dict from string arg_name to a list of values,
         possibly empty, which should not elicit a warning.
-      arg_spec: Output from inspect.getargspec on the called function.
+      arg_spec: Output from tf_inspect.getargspec on the called function.
 
     Returns:
       Dictionary from arg_name to DeprecatedArgSpec.
@@ -209,7 +221,7 @@ def deprecated_args(date, instructions, *deprecated_arg_names_or_tuples):
     decorator_utils.validate_callable(func, 'deprecated_args')
     deprecated_arg_names = _get_arg_names_to_ok_vals()
 
-    arg_spec = inspect.getargspec(func)
+    arg_spec = tf_inspect.getargspec(func)
     deprecated_positions = _get_deprecated_positional_arguments(
         deprecated_arg_names, arg_spec)
 
@@ -255,33 +267,36 @@ def deprecated_args(date, instructions, *deprecated_arg_names_or_tuples):
     @functools.wraps(func)
     def new_func(*args, **kwargs):
       """Deprecation wrapper."""
-      invalid_args = []
-      named_args = inspect.getcallargs(func, *args, **kwargs)
-      for arg_name, spec in iter(deprecated_positions.items()):
-        if (spec.position < len(args) and
-            not (spec.has_ok_value and
-                 _same_value(named_args[arg_name], spec.ok_value))):
-          invalid_args.append(arg_name)
-      if is_varargs_deprecated and len(args) > len(arg_spec.args):
-        invalid_args.append(arg_spec.varargs)
-      if is_kwargs_deprecated and kwargs:
-        invalid_args.append(arg_spec.keywords)
-      for arg_name in deprecated_arg_names:
-        if (arg_name in kwargs and
-            not (deprecated_positions[arg_name].has_ok_value and
-                 _same_value(named_args[arg_name],
-                             deprecated_positions[arg_name].ok_value))):
-          invalid_args.append(arg_name)
-      for arg_name in invalid_args:
-        logging.warning(
-            'From %s: calling %s (from %s) with %s is deprecated and will '
-            'be removed after %s.\nInstructions for updating:\n%s',
-            _call_location(), decorator_utils.get_qualified_name(func),
-            func.__module__, arg_name, date, instructions)
+      if _PRINT_DEPRECATION_WARNINGS:
+        invalid_args = []
+        named_args = tf_inspect.getcallargs(func, *args, **kwargs)
+        for arg_name, spec in iter(deprecated_positions.items()):
+          if (spec.position < len(args) and
+              not (spec.has_ok_value and
+                   _same_value(named_args[arg_name], spec.ok_value))):
+            invalid_args.append(arg_name)
+        if is_varargs_deprecated and len(args) > len(arg_spec.args):
+          invalid_args.append(arg_spec.varargs)
+        if is_kwargs_deprecated and kwargs:
+          invalid_args.append(arg_spec.keywords)
+        for arg_name in deprecated_arg_names:
+          if (arg_name in kwargs and
+              not (deprecated_positions[arg_name].has_ok_value and
+                   _same_value(named_args[arg_name],
+                               deprecated_positions[arg_name].ok_value))):
+            invalid_args.append(arg_name)
+        for arg_name in invalid_args:
+          logging.warning(
+              'From %s: calling %s (from %s) with %s is deprecated and will '
+              'be removed %s.\nInstructions for updating:\n%s',
+              _call_location(), decorator_utils.get_qualified_name(func),
+              func.__module__, arg_name,
+              'in a future version' if date is None else ('after %s' % date),
+              instructions)
       return func(*args, **kwargs)
-    new_func.__doc__ = _add_deprecated_arg_notice_to_docstring(
-        func.__doc__, date, instructions)
-    return new_func
+    return tf_decorator.make_decorator(func, new_func, 'deprecated',
+                                       _add_deprecated_arg_notice_to_docstring(
+                                           func.__doc__, date, instructions))
   return deprecated_wrapper
 
 
@@ -295,6 +310,7 @@ def deprecated_arg_values(date, instructions, **deprecated_kwargs):
     will be removed after <date>. Instructions for updating:
       <instructions>
 
+  If `date` is None, 'after <date>' is replaced with 'in a future version'.
   <function> will include the class name if it is a method.
 
   It also edits the docstring of the function: ' (deprecated arguments)' is
@@ -302,8 +318,8 @@ def deprecated_arg_values(date, instructions, **deprecated_kwargs):
   prepended to the rest of the docstring.
 
   Args:
-    date: String. The date the function is scheduled to be removed. Must be
-      ISO 8601 (YYYY-MM-DD).
+    date: String or None. The date the function is scheduled to be removed.
+      Must be ISO 8601 (YYYY-MM-DD), or None
     instructions: String. Instructions on how to update code using the
       deprecated function.
     **deprecated_kwargs: The deprecated argument values.
@@ -312,7 +328,8 @@ def deprecated_arg_values(date, instructions, **deprecated_kwargs):
     Decorated function or method.
 
   Raises:
-    ValueError: If date is not in ISO 8601 format, or instructions are empty.
+    ValueError: If date is not None or in ISO 8601 format, or instructions are
+      empty.
   """
   _validate_deprecation_args(date, instructions)
   if not deprecated_kwargs:
@@ -324,18 +341,21 @@ def deprecated_arg_values(date, instructions, **deprecated_kwargs):
     @functools.wraps(func)
     def new_func(*args, **kwargs):
       """Deprecation wrapper."""
-      named_args = inspect.getcallargs(func, *args, **kwargs)
-      for arg_name, arg_value in deprecated_kwargs.items():
-        if arg_name in named_args and named_args[arg_name] == arg_value:
-          logging.warning(
-              'From %s: calling %s (from %s) with %s=%s is deprecated and will '
-              'be removed after %s.\nInstructions for updating:\n%s',
-              _call_location(), decorator_utils.get_qualified_name(func),
-              func.__module__, arg_name, arg_value, date, instructions)
+      if _PRINT_DEPRECATION_WARNINGS:
+        named_args = tf_inspect.getcallargs(func, *args, **kwargs)
+        for arg_name, arg_value in deprecated_kwargs.items():
+          if arg_name in named_args and named_args[arg_name] == arg_value:
+            logging.warning(
+                'From %s: calling %s (from %s) with %s=%s is deprecated and '
+                'will be removed %s.\nInstructions for updating:\n%s',
+                _call_location(), decorator_utils.get_qualified_name(func),
+                func.__module__, arg_name, arg_value,
+                'in a future version' if date is None else ('after %s' % date),
+                instructions)
       return func(*args, **kwargs)
-    new_func.__doc__ = _add_deprecated_arg_notice_to_docstring(
-        func.__doc__, date, instructions)
-    return new_func
+    return tf_decorator.make_decorator(func, new_func, 'deprecated',
+                                       _add_deprecated_arg_notice_to_docstring(
+                                           func.__doc__, date, instructions))
   return deprecated_wrapper
 
 
@@ -363,3 +383,13 @@ def deprecated_argument_lookup(new_name, new_value, old_name, old_value):
 def rewrite_argument_docstring(old_doc, old_argument, new_argument):
   return old_doc.replace('`%s`' % old_argument, '`%s`' % new_argument).replace(
       '%s:' % old_argument, '%s:' % new_argument)
+
+
+@tf_contextlib.contextmanager
+def silence():
+  """Temporarily silence deprecation warnings."""
+  global _PRINT_DEPRECATION_WARNINGS
+  print_deprecation_warnings = _PRINT_DEPRECATION_WARNINGS
+  _PRINT_DEPRECATION_WARNINGS = False
+  yield
+  _PRINT_DEPRECATION_WARNINGS = print_deprecation_warnings
